/*
 * Copyright (c) 1999-2014, Ecole des Mines de Nantes
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Ecole des Mines de Nantes nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE REGENTS AND CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

// Raphaël Boudreault/Improved CP-Based Lagrangian Relaxation Approach with an Application to the TSP (IJCAI 2021)

package TSP;

import org.chocosolver.graphsolver.GraphModel;
import org.chocosolver.graphsolver.cstrs.cost.tsp.lagrangianRelaxation.improvedAlgorithms.configOptions;
import org.chocosolver.graphsolver.search.strategy.GraphSearch;
import org.chocosolver.graphsolver.variables.UndirectedGraphVar;
import org.chocosolver.solver.Model;
import org.chocosolver.solver.Solver;
import org.chocosolver.solver.variables.IntVar;
import org.chocosolver.util.objects.graphs.UndirectedGraph;
import org.chocosolver.util.objects.setDataStructures.SetType;

import java.util.Locale;

/**
 * Solves the Traveling Salesman Problem
 * Parses TSP instances of the TSPLIB library
 * See <a href = "http://comopt.ifi.uni-heidelberg.de/software/TSPLIB95/">TSPLIB</a>
 * <p/>
 *
 * This is an exact approach dedicated to prove optimality of a solution.
 * It is assumed that a local search (e.g. LKH) algorithm has been performed
 * as a pre-processing step
 *
 * @author Jean-Guillaume Fages
 * @since Oct. 2012
 */
public class TSP {

    //***********************************************************************************
    // MAIN
    //***********************************************************************************

    public static void main(String[] args) {
		if (args.length != 3) {
			System.out.println("Usage : java -jar TSP_main.jar <Instance directory> <Instance name (_.tsp)> <Path to .json config>");
			return;
		}
        Locale.setDefault(new Locale("en", "US"));
        String REPO = args[0];
		String INSTANCE = args[1];
		String JSONfile = args[2];
        int[][] data = TSP_Utils.parseInstance(REPO+"/"+INSTANCE+".tsp", 500);
        int upperBound = TSP_Utils.getOptimum(INSTANCE,REPO+"/UB.csv");
        configOptions options = new configOptions(JSONfile);

        new TSP(data, upperBound, options);
    }

    //***********************************************************************************
    // SOLVER
    //***********************************************************************************

    public TSP(int[][] costMatrix, int initialUB, configOptions options){
        super();
        final int n = costMatrix.length;
        int LIMIT = 1800; // in seconds

        GraphModel model = new GraphModel("TSP");
        // variables
        IntVar totalCost = model.intVar("obj", 0, initialUB, true);
        // creates a graph containing n nodes
        UndirectedGraph GLB = new UndirectedGraph(model, n, SetType.LINKED_LIST, true);
        UndirectedGraph GUB = new UndirectedGraph(model, n, SetType.BIPARTITESET, true);
        // adds potential edges
        for (int i = 0; i < n; i++) {
            for (int j = i + 1; j < n; j++) {
                GUB.addEdge(i, j);
            }
        }
        UndirectedGraphVar graph = model.graphVar("G", GLB, GUB);

        // constraints (TSP basic model + lagrangian relaxation)
        model.tsp(graph, totalCost, costMatrix, 1).post();

        Solver solver = model.getSolver();

        // Set search heuristic
        solver.setSearch(new GraphSearch(graph, costMatrix).configure(GraphSearch.MAX_COST).useLastConflict());
        //solver.setSearch(new GraphSearch(graph, costMatrix).configure(GraphSearch.MAX_COST));
        //solver.setSearch(new GraphSearch(graph, costMatrix).configure(GraphSearch.LEX));

        // Set time limit
        solver.limitTime(LIMIT+"s");

        model.setObjective(Model.MINIMIZE,totalCost);
        options.printConfig();

        while (solver.solve())
        {
            //solver.printShortStatistics();
            System.out.println("New solution found: " + totalCost);
            //printSolution(graph);
        }
        if (solver.getTimeCount() < LIMIT){
            solver.printShortStatistics();
        } else {
            System.out.println("TIMEOUT (no optimality proof)");
        }
    }

    public void printSolution(UndirectedGraphVar graph) {
        int i = 0;
        int last_i = -1;
        do {
            System.out.print(i + ", ");
            var nei = graph.getMandNeighOf(i).newIterator();
            int temp = i;
            i = nei.next();
            if (i == last_i) i = nei.next();
            last_i = temp;
        } while (i != 0);
        System.out.print(0);
    }
}